let platform = process.platform;

if (platform === "win32") {
  // Import dependencies
  const koffi = require("koffi");
  let path = require("path");
  let fs = require("fs");

  let JLinkDrv, JLinkDriverffiMapping;

  // Load the STJlink dll function mapping with Koffi
  STJLinkLibrary = path.join(__dirname, "STJlinkLibrary.dll");
  const JLinkLib = koffi.load(STJLinkLibrary);
  JLinkDriverffiMapping = {
    opendevice: JLinkLib.func("string opendevice(string deviceName)"),
    getString: JLinkLib.func("string getString (void)"),
    closejlink: JLinkLib.func("int closejlink (void)"),
    readaddress: JLinkLib.func("string readaddress (string address)"),
    writeaddress: JLinkLib.func("int writeaddress (string address, int value)"),
    writeaddressfloat: JLinkLib.func("int writeaddressfloat (string address, float value)"),
    writeaddressdouble: JLinkLib.func("int writeaddressdouble (string address, double value)"),
    showjlink: JLinkLib.func("int showjlink (void)"),
    printjlink: JLinkLib.func("int printjlink (void)")
  };

  JLinkDrv = (function () {
    let map = {};
    for (let [key, value] of Object.entries(JLinkDriverffiMapping)) {
      /**
       * Create function to add log before original function call.
       * @param {...any} args - Mapped functions and arg.
       */
      let callDriverAPI = (...args) => {
        //console.log("key", key);
        //console.log("args", args);
        return value.apply(this, args);
      };
      //property writable set to true for unitary test purpose
      map = Object.defineProperty(map, key, {
        value: callDriverAPI,
        writable: true
      });
    }
    return map;
  })();

  const TYPE_U8 = 1;
  const TYPE_S8 = 2;
  const TYPE_U16 = 3;
  const TYPE_S16 = 4;
  const TYPE_U32 = 5;
  const TYPE_S32 = 6;
  const TYPE_U64 = 7; // Not implemented because not supported in node.js
  const TYPE_S64 = 8; // Not implemented because not supported in node.js
  const TYPE_FLOAT = 9;
  const TYPE_DOUBLE = 10;

  /**
   * Function to reverse a 4 bytes string.
   * @param {*} hex - A 4 bytes string in hexadecimal.
   */
  function reverseHexString(hex) {
    return hex.slice(6, 8) + hex.slice(4, 6) + hex.slice(2, 4) + hex.slice(0, 2);
  }

  /**
   * Function to calculate a float from a hexadecimal value.
   * @param {*} value - An hexadecimal value.
   */
  function hex2float(value) {
    //create a binary array from the address
    let binaryArray = new Uint8Array(value.match(/.{1,2}/g).map(byte => parseInt(byte, 16)));

    // Create a dataview from the array
    let dataView = new DataView(binaryArray.buffer);

    let floatValue = dataView.getFloat32(0);

    return floatValue;
  }

  /**
   * Function to calculate a Signed 32 bits number from a hexadecimal value.
   * @param {*} value - An hexadecimal value.
   */
  function hex2s32(value) {
    //create a binary array from the address
    let binaryArray = new Uint8Array(value.match(/.{1,2}/g).map(byte => parseInt(byte, 16)));

    // Create a dataview from the array
    let dataView = new DataView(binaryArray.buffer);

    let s32Value = dataView.getInt32(0);

    return s32Value;
  }

  /**
   * Function to calculate a Unsigned 8 bits number from a hexadecimal value.
   * @param {*} value - An hexadecimal value.
   */
  function hex2u8(value) {
    //create a binary array from the address
    let lastByte = value.substr(-2); // // extract last byte of the hexadecimal string
    let u8Value = parseInt(lastByte, 16);

    return u8Value;
  }

  /**
   * Function to calculate a Signed 8 bits number from a hexadecimal value.
   * @param {*} value - An hexadecimal value.
   */
  function hex2s8(value) {
    //create a binary array from the address
    let lastByte = value.substr(-2); // extract last byte of the hexadecimal string
    let s8Value = new Int8Array([parseInt(lastByte, 16)]);

    return s8Value;
  }

  /**
   * Function to calculate a Unsigned 16 bits number from a hexadecimal value.
   * @param {*} value - An hexadecimal value.
   */
  function hex2u16(value) {
    //create a binary array from the address
    let binaryArray = new Uint8Array(value.match(/.{1,2}/g).map(byte => parseInt(byte, 16)));

    // Create a dataview from the array
    let dataView = new DataView(binaryArray.buffer);
    let u16Value = dataView.getUint16(2);

    return u16Value;
  }

  /**
   * Function to calculate a Signed 16 bits number from a hexadecimal value.
   * @param {*} value - An hexadecimal value.
   */
  function hex2s16(value) {
    let lastTwoBytes = value.substr(-4); // // extract the two last bytes of the hexadecimal string
    let int16 = new Int16Array([parseInt(lastTwoBytes, 16)]); // convert the two last bytes to 16 bits signed
    return int16;
  }

  /**
   * Function to calculate a Double number from an address.
   * @param {*} address - An hexadecimal address as the base for the location of the doublenumber on 8 bytes.
   * @param {*} index - The beginning index in the table.
   */
  function doubleFromAddress(address, index) {
    let v1 = address[index];
    let v2 = address[index + 1];

    return Buffer.from(reverseHexString(v1) + reverseHexString(v2), "hex").readDoubleLE(0);
  }

  /**
   * Function to read memory group
   * The jlink probe will read the data requested.
   *
   * @param {*} myData  - Info about data to read.
   */
  function JlinkReadMemoryGroup(myData) {
    let res = {
      result: 1,
      data: []
    };
    let numItems = 0;
    for (var i = 0; i < myData.length; i++) {
      if (myData[i].nb != undefined) {
        numItems = numItems + myData[i].nb;
      }
      if (numItems == 0) {
        numItems = 1;
      }
      if (numItems > 0) {
        if (myData[i].type == TYPE_DOUBLE) {
          numItems++;
        }

        const addressToRead = "0x" + parseInt(myData[i].address, 16).toString(16).toUpperCase() + " " + numItems;
        let result = JLinkDrv.readaddress(addressToRead);
        res.data.push([]);
        let resArray = result.split(" ");
        for (let j = 0; j < numItems; j++) {
          switch (myData[i].type) {
            case TYPE_U8: //1
              yValue = hex2u8(resArray[j]);
              break;
            case TYPE_S8: //2
              yValue = hex2s8(resArray[j])[0];
              break;
            case TYPE_U16: //3
              yValue = hex2u16(resArray[j]);
              break;
            case TYPE_S16: //4
              yValue = hex2s16(resArray[j])[0];
              break;
            case TYPE_U32: //5
              yValue = parseInt(resArray[j], 16);
              break;
            case TYPE_S32: //6
              yValue = hex2s32(resArray[j]);
              break;
            case TYPE_FLOAT: //9
              yValue = hex2float(resArray[j]);
              break;
            case TYPE_DOUBLE: //10
              yValue = doubleFromAddress(resArray, j);
              j++;
              break;
            default:
              yValue = parseInt(resArray[j], 16);
          }

          let timestamp = +new Date();
          res.data[i].push({
            y: yValue,
            x: timestamp
          });
        }
      }
      numItems = 0;
    }
    return res;
  }

  /**
   * Function to write memory group
   * The jlink probe will write the data requested.
   *
   * @param {*} myData  - Info about data to write.
   */
  function JlinkWriteMemoryGroup(myData) {
    let res = {
      result: 1,
      data: []
    };
    let listOfVariable;
    if (myData.variablelist != undefined) {
      listOfVariable = myData.variablelist;
    } else {
      listOfVariable = myData;
    }
    for (var i = 0; i < listOfVariable.length; i++) {
      const addressToWrite = listOfVariable[i].address.toUpperCase();
      value = parseInt(listOfVariable[i].value);
      switch (listOfVariable[i].type) {
        case TYPE_U8: //1
          JLinkDrv.writeaddress(addressToWrite, value);
          break;
        case TYPE_S8: //2
          JLinkDrv.writeaddress(addressToWrite, value);
          break;
        case TYPE_U16: //3
          JLinkDrv.writeaddress(addressToWrite, value);
          break;
        case TYPE_S16: //4
          let signed16bitValue = (value << 16) >> 16;
          JLinkDrv.writeaddress(addressToWrite, signed16bitValue);
          break;
        case TYPE_U32: //5
          JLinkDrv.writeaddress(addressToWrite, value);
          break;
        case TYPE_S32: //6
          JLinkDrv.writeaddress(addressToWrite, value);
          break;
        case TYPE_FLOAT: //9
          value = parseFloat(listOfVariable[i].value);
          JLinkDrv.writeaddressfloat(addressToWrite, value);
          break;
        case TYPE_DOUBLE: //10
          value = parseFloat(listOfVariable[i].value);
          JLinkDrv.writeaddressdouble(addressToWrite, value);
          break;
        default:
          JLinkDrv.writeaddress(addressToWrite, value);
      }
    }

    return res;
  }

  /**
   * Function to read the list of supported MCU target.
   * @returns {[string]} Models List of MCU or "File not Found" in first element of the list.
   */
  function readJlinkSupportedHW() {
    const models = [];
    // check if file is present
    const filename = path.join(__dirname, "../SupportedHW.txt");
    if (!fs.existsSync(filename)) {
      const model = "File not found";
      models.push(model);
      return models;
    }

    // read the file
    const data = fs.readFileSync(filename, "utf8");
    const lines = data.split("\n");

    // filter the file to keep only STM32
    for (let i = 0; i < lines.length - 1; i++) {
      const line = lines[i];
      const fields = line.split(",");
      const vendorField = fields[0];

      if (vendorField === '"ST"') {
        const modelField = fields[1];
        const replacetrimresult = modelField.replace(/"/g, "").trim();
        if (replacetrimresult != null) {
          const regexresult = replacetrimresult.match(/^[\w\d]+/);
          if (regexresult != null) {
            const model = regexresult[0];
            if (model.startsWith("STM")) {
              if (!replacetrimresult.includes(" ")) {
                models.push(model);
              }
            }
          }
        }
      }
    }
    return models;
  }

  // module exports
  module.exports.printjlink = JLinkDrv.printjlink;
  module.exports.showjlink = JLinkDrv.showjlink;
  module.exports.opendevice = JLinkDrv.opendevice;
  module.exports.readaddress = JLinkDrv.readaddress;
  module.exports.writeaddress = JLinkDrv.writeaddress;
  module.exports.closejlink = JLinkDrv.closejlink;
  module.exports.getString = JLinkDrv.getString;

  module.exports.readJlinkSupportedHW = readJlinkSupportedHW;
  module.exports.JlinkReadMemoryGroup = JlinkReadMemoryGroup;
  module.exports.JlinkWriteMemoryGroup = JlinkWriteMemoryGroup;
}
